#DEFINE DB_TRANSMANUAL 2
DEFINE CLASS LOCKS AS CUSTOM

  ** cNetworkMachine  User and Workstation name
  PROTECTED cNetworkMachine
  cNetworkMachine = SYS(0)

  ** SQL Connection Handler
  PROTECTED nSqlHandler
  nSqlHandler   = -1

  ** Connection String used to connect to the SQL Server
  PROTECTED cSqlString
  cSqlString=[Driver={SQL Server};Server=Boris;Trusted_Connection=yes;DataBase=Test;]

  ** SQL Server Table Name used for the Locking
  PROTECTED cSQLServerTableName
  cSQLServerTableName = [Locks]

  ** Name of the cursor where we receive results
  PROTECTED cUniqueCursorname
  cUniqueCursorname = SYS(2015)


  PROCEDURE LockTable(lcAliasName, bForce)
    *** Tries to lock the entire table
    *** The second parameter skips the search, and forces
    *** the alias name insertion into the Locks Table
    *** This method simulate FLOCK () function
    RETURN THIS.LockUnLockTable(lcAliasName, bForce, .F.)
  ENDPROC

  PROCEDURE UnLockTable(lcAliasName, bForce)
    *** Unlocks the table
    THIS.LockUnLockTable(lcAliasName, bForce, .T.)
  ENDPROC


  PROCEDURE LockRecords(lcRecordNumberList, lcAliasName, lcKeyFields)

    *** This method simulates
    *** the RLOCK() function.
    *** The parameters are:
    *** lcRecordNumberList -string contained list of the records to be LOCKed
    *** lcAliasName  Alias name where we try to Lock these records
    *** lcKeyFields  - The fields list which values makes a record unique
    RETURN THIS.LockUnlockRecords(lcRecordNumberList, lcAliasName,;
      lcKeyFields, .F., PCOUNT())
  ENDPROC

  PROCEDURE UnLockRecords(lcRecordNumberList, lcAliasName, lcKeyFields)
    *** Unlocks Records
    THIS.LockUnlockRecords(lcRecordNumberList, lcAliasName, lcKeyFields, .T., PCOUNT())
  ENDPROC


  PROTECTED PROCEDURE SQLUnLock(eExpression)
    TEXT TO lcSql NOSHOW  PRETEXT 15 TEXTMERGE
                DELETE FROM <<this.cSQLServerTableName>>
                 WHERE LEFT(Expression,<<LEN(m.eExpression)>>) = '<<ALLTRIM(m.eExpression)>>'
                  AND WorkStation = '<<this.cNetworkMachine>>'
    ENDTEXT

    SQLEXEC(THIS.nSqlHandler, lcSql)
    TEXT TO lcSql NOSHOW  PRETEXT 15 TEXTMERGE
                SELECT COUNT(*) FROM <<this.cSQLServerTableName>> WITH (NOLOCK)
                       WHERE WorkStation = '<<this.cNetworkMachine>>'
    ENDTEXT


    ** Check to see if table contains some records added from this WorkStation
    ** if not, close connection.
    IF SQLEXEC(THIS.nSqlHandler, lcSql, THIS.cUniqueCursorname) > 0
      IF RECCOUNT(THIS.cUniqueCursorname) == 0
        SQLDISCONNECT(THIS.nSqlHandler)
        THIS.nSqlHandler = -1
      ENDIF
      USE IN SELECT(THIS.cUniqueCursorname)
    ENDIF
    RETURN
  ENDPROC


  PROTECTED PROCEDURE LockUnlockRecords(lcRecordNumberList, lcAliasName, lcKeyFields,;
      bUnLock, m.lnPcount)
    LOCAL ARRAY larecords[1],laKeys[1]
    LOCAL lnFor, lnLines, lcAlias, lnPcount, lbRecordList, lcExpr, lcSql
    LOCAL lcCurrentAlias, leFieldvalue
    LOCAL lbIsCursorAdapter, loCursorAdapter, lcTableName
    STORE [] TO larecords, laKeys


    *** Check for valid connection
    IF NOT THIS.ConnectSql()
      RETURN .F.
    ENDIF

    *** If this class is created on Apllication level and forms used Private DataSession
    *** Alias opened in such a form is not reachable
    IF TYPE([_screen.ActiveForm]) == [O] AND;
        NOT ISNULL(_SCREEN.ACTIVEFORM)
      SET DATASESSION TO (_SCREEN.ACTIVEFORM.DATASESSIONID)
    ENDIF

    m.lcCurrentAlias = ALIAS()
    m.lbRetVal       = .T.

    IF m.lnPcount = 0
      m.lcAlias  = m.lcCurrentAlias
    ELSE
      m.lcAlias  = m.lcRecordNumberList
    ENDIF

    IF VARTYPE(m.lcAlias) == [C] AND;
        SELECT(m.lcAlias) = 0     AND;
        m.lnPcount > 1
      m.lcAlias = m.lcAliasName
    ENDIF

    IF VARTYPE(m.lcAlias) # [C] OR;
        SELECT(m.lcAlias) = 0
      m.lcAlias = m.lcCurrentAlias
    ENDIF

    IF SELECT(m.lcAlias) = 0
      RETURN .F.
    ENDIF

    m.lbIsCursorAdapter = .T. && Specified Alias, belongs to CursorAdapter
    TRY
      loCursorAdapter = GETCURSORADAPTER(m.lcAlias)
    CATCH
      m.lbIsCursorAdapter = .F. && Nope, no cursoradapter
    ENDTRY

    IF m.lbIsCursorAdapter
      m.lcExpr = loCursorAdapter.TABLES
    ELSE
      m.lcExpr = CURSORGETPROP("Tables",m.lcAlias)
    ENDIF

    m.lcTableName = m.lcAlias
    IF NOT EMPTY(m.lcExpr)
      m.lcTableName = m.lcExpr
    ENDIF

    IF NOT m.bUnLock

      ** Check to see if the table is not already Locked
      TEXT TO lcSql NOSHOW  PRETEXT 15 TEXTMERGE
                   SELECT * FROM <<this.cSQLServerTableName>> WITH (NOLOCK)
                            WHERE Expression= '<<[_]+m.lcTableName+[_]>>'
      ENDTEXT

      IF SQLEXEC(THIS.nSqlHandler,m.lcSql,THIS.cUniqueCursorname) < 0
        m.lbRetVal = .F.
      ENDIF

      IF m.lbRetVal AND RECCOUNT(THIS.cUniqueCursorname) > 0
        m.lbRetVal = .F.
      ENDIF

      USE IN SELECT(THIS.cUniqueCursorname)
      IF NOT EMPTY(m.lcCurrentAlias)
        SELECT (m.lcCurrentAlias)
      ENDIF
    ENDIF

    IF m.lbRetVal
      IF m.lnPcount > 0 AND VARTYPE(m.lcRecordNumberList) == [C]
        lnLines = ALINES(larecords ,m.lcRecordNumberList,1+2+4,[,])
        IF lnLines = 0 OR TYPE(larecords[1]) # [N]
          larecords[1] = TRANSFORM(RECNO(m.lcAlias))
        ENDIF
      ELSE
        larecords[1] = TRANSFORM(RECNO(m.lcAlias))
      ENDIF

      IF m.lnPcount < 3 OR VARTYPE(m.lcKeyFields) # [C]
        IF m.lbIsCursorAdapter
          m.lcKeyFields = loCursorAdapter.KEYFIELDLIST
        ELSE
          m.lcKeyFields = CURSORGETPROP("KeyFieldList", m.lcAlias)
        ENDIF
      ENDIF

      m.lnLines = ALINES(laKeys,m.lcKeyFields,1+2+4,[,])
      IF m.lnLines = 0
        DIMENSION laKeys[FCOUNT(m.lcAlias)]
        FOR m.lnFor = 1 TO FCOUNT(m.lcAlias)
          laKeys[m.lnFor] = FIELD(m.lnFor,m.lcAlias)
        NEXT
      ENDIF

      IF NOT m.bUnLock
        lcnameTran = SYS(2015)
        TEXT TO lcSql NOSHOW  PRETEXT 15 TEXTMERGE
                      BEGIN TRANSACTION
                      SAVE  TRANSACTION <<lcnameTran>>
        ENDTEXT
        SQLEXEC(THIS.nSqlHandler,lcSql)
      ENDIF

      m.lnCurrentRecord = RECNO(m.lcAlias)
      FOR m.lnFor = 1 TO ALEN(larecords,1)
        GO VAL(larecords[m.lnFor]) IN (m.lcAlias)
        STORE [] TO m.lcExpr
        FOR m.lnLines = 1 TO ALEN(laKeys,1)
          m.leFieldvalue = EVALUATE(m.lcAlias+[.]+laKeys[m.lnLines])
          m.lcExpr       = m.lcExpr + NVL(TRANSFORM(m.leFieldvalue),[])
        NEXT
        IF m.bUnLock
          THIS.SQLUnLock([_]+m.lcTableName+[_]+m.lcExpr)
          m.lbRetVal = .T.  && Just to Skip LastUnlock
        ELSE
          m.lbRetVal = m.lbRetVal AND THIS.SQLLock([_]+m.lcTableName+[_]+m.lcExpr)
        ENDIF
      NEXT
      GOTO m.lnCurrentRecord IN m.lcAlias

      IF NOT m.bUnLock
        IF NOT m.lbRetVal
          TEXT TO lcSql NOSHOW  PRETEXT 15 TEXTMERGE
                         ROLLBACK TRANSACTION <<lcnameTran>>
          ENDTEXT
        ELSE
          TEXT TO lcSql NOSHOW  PRETEXT 15 TEXTMERGE
                         COMMIT TRANSACTION <<lcnameTran>>
          ENDTEXT
        ENDIF
        SQLEXEC(THIS.nSqlHandler,lcSql)
      ENDIF
    ENDIF
    RETURN m.lbRetVal
  ENDPROC

  PROTECTED PROCEDURE LockUnLockTable(lcAliasName, bForce, bUnLock)
    LOCAL lcTableName, m.RetVal
    LOCAL lbIsCursorAdapter, loCursorAdapter

    m.RetVal = .T.
    IF NOT m.bForce
      *** Locking table
      *** because the Alias of the table is not reliable I try to get real Table name
      *** bForce = .t. - means that the exact SQL Server table name is passed to the
      ***                Method, so skip the check for Table property of the cursor

      m.lcAliasName       = IIF(VARTYPE(m.lcAliasName) == [C], m.lcAliasName, ALIAS())
      m.lbIsCursorAdapter = .T. && Specified Alias, belongs to CursorAdapter
      TRY
        loCursorAdapter = GETCURSORADAPTER(m.lcAliasName)
      CATCH
        m.lbIsCursorAdapter = .F. && Nope, Alias is not defined with CursordAdapter
      ENDTRY

      TRY
        IF m.lbIsCursorAdapter
          m.lcTableName = loCursorAdapter.TABLES
        ELSE
          m.lcTableName = CURSORGETPROP("Tables",m.lcAliasName)
        ENDIF
        IF NOT EMPTY(m.lcTableName)
          ** If this is empty no table name is specified in CA or View
          m.lcAliasName = m.lcTableName
        ENDIF
      CATCH
        *** Try to Lock/UnLock NATIVE VFP Table
        m.RetVal = .F.
      ENDTRY
    ENDIF  &&NOT m.bForce

    IF m.RetVal
      IF m.bUnLock
        m.RetVal = THIS.SQLUnLock([_]+m.lcAliasName+[_])
      ELSE
        m.RetVal = THIS.SQLLock([_]+m.lcAliasName+[_])
      ENDIF
    ENDIF
    RETURN m.RetVal
  ENDPROC

  PROCEDURE DESTROY
    IF THIS.nSqlHandler > 0
      SQLDISCONNECT(THIS.nSqlHandler)
      THIS.nSqlHandler = -1
    ENDIF
  ENDPROC

  PROTECTED PROCEDURE SQLLock(eExpression, bDataBaseLock, bDataBaseExclusive)

    LOCAL lbReturnValue, lcCurrentAlias, lcUniqueTrans

    ** Check for valid connection
    lcCurrentAlias = ALIAS()
    IF NOT THIS.ConnectSql()
      RETURN .F.
    ENDIF

    m.lbReturnValue = .T.
    * Trying to lock Record or Table
    TEXT TO lcSql NOSHOW  PRETEXT 15 TEXTMERGE
                INSERT INTO <<this.cSQLServerTableName>> (Expression, WorkStation);
                 VALUES ('<<LEFT(m.eExpression,200)>>','<<this.cNetworkMachine>>')
    ENDTEXT

    IF SQLEXEC(THIS.nSqlHandler, lcSql) < 0
      m.lbReturnValue = .F.
    ENDIF

    USE IN SELECT(THIS.cUniqueCursorname)
    IF NOT EMPTY(m.lcCurrentAlias)
      SELECT (m.lcCurrentAlias)
    ENDIF

    RETURN m.lbReturnValue

  ENDPROC

  PROTECTED PROCEDURE ConnectSql
    *** This method check for valid connection
    *** if there is no such a connection
    *** reconnect again
    LOCAL lbRetVal
    m.lbRetVal = (THIS.nSqlHandler > 0)
    IF THIS.nSqlHandler > 0
      TRY
        *** Check to see if the connection is still valid
        SQLGETPROP(THIS.nSqlHandler,"ConnectBusy")
      CATCH
        *** Something is happened with the connection
        m.lbRetVal = .F.
      ENDTRY
    ENDIF

    IF NOT m.lbRetVal
      THIS.nSqlHandler = SQLSTRINGCONNECT(THIS.cSqlString)

      IF THIS.nSqlHandler < 1
        THIS.nSqlHandler = -1
        RETURN .F.
      ENDIF

      *** Set to manual transactions
      SQLSETPROP(THIS.nSqlHandler,"Transactions"      ,DB_TRANSMANUAL)

      *** If the sudden disconnect happens all changes must be
      *** ROLLBACK
      SQLSETPROP(THIS.nSqlHandler,"DisconnectRollback", .T.)

      *** 2 seconds for query
      SQLSETPROP(THIS.nSqlHandler,"QueryTimeOut"      , 2)

      *** Starting manual transaction
      SQLEXEC(THIS.nSqlHandler,[BEGIN TRANSACTION])
    ENDIF
    RETURN .T.
  ENDPROC


  PROCEDURE DESTROY
    IF THIS.nSqlHandler > 0
      SQLDISCONNECT(THIS.nSqlHandler)
    ENDIF
  ENDPROC
ENDDEFINE
